#include <stdio.h>
#include "pairwise_cyclic_alignment.h"
#include "average_linkage_cluster.h"
#include "fasta.h"
#include "cycl_util.h"
#include <ostream>
#include <iostream>
#include <fstream>
#include <limits.h>
#include <cmath>

bool linear_alignment;
bool local_alignment;
bool fast_heuristic;
bool multiple;
bool tree_guided;
bool tree_weights;
std::string output_filename;
float gap_open_penalty;
float gap_ext_penalty;
bool gap_cost_changed;
bool get_shifted_sequences;
bool use_joint_match_score;
bool verbose;
bool clustalw_output;

int num_of_options;

int header_width;
int body_width;

typedef pairwise_cyclic_alignment::sequence_type sequence_type;
typedef std::vector<sequence_type> sequence_vector;
typedef average_linkage_cluster guide_tree_alignment;

// =========================================================	
void print_help_message()
// =========================================================	
{
	std::cout<<"\nusage: cyclope [options] fastafile\n\n";
	std::cout<<"options:\n";
	std::cout<<" -a   direct agglomorative alignment without guide tree (evtl. slow)\n";
	std::cout<<" -v   verbose mode\n";
	std::cout<<" -p   pairwise alignment only regarding the first two sequences in the fasta file\n";
	std::cout<<" -H   fast heuristic pairwise alignment only regarding the first two sequences in the fasta file\n";
	std::cout<<" -l   pairwise linear (non-cyclic) alignment only regarding the first two sequences in the fasta file\n";
	std::cout<<" -o   pairwise local (non-cyclic) alignment only regarding the first two sequences in the fasta file\n";
	std::cout<<" -s   output shifted sequences in fasta format\n";
	std::cout<<" -C   produce pseudo-ClustalW output for SplitsTree input\n";
	std::cout<<" -j   use joint match score\n";
	std::cout<<" -g <gap_open_penalty> <gap_ext_penalty>   set gap costs\n";
	std::cout<<" -L <header_w> <body_w>   format output\n";
	std::cout<<" -O <filename> write output to <filename> instead of stdout\n";
   std::cout<<"\n";
}

// =========================================================	
int read_parameters(int argc, char** argv)
// =========================================================	
{
	linear_alignment = false;
	local_alignment = false;
	multiple = true;
	tree_guided = true;
	tree_weights = false;
	header_width = 8;
	body_width = 60;
	output_filename = "";
	gap_cost_changed = false;
	get_shifted_sequences = false;
	use_joint_match_score = false;
	pairwise_cyclic_alignment::use_joint_match_score = false;
	verbose = false;
	
	if (argc<2)
	{
		return false;
	}

	num_of_options = 1;
	
	int i;
	for (i=1; i<argc && argv[i][0]=='-'; i++)
	{
		num_of_options++;
		switch (argv[i][1])
		{
			case 'a' : {
				if (strlen(argv[i])>2)
				{
					std::cerr<<"Unknown option: "<<argv[i]<<"\n";
					return false;
				}
				tree_guided = false;
				break;
			}
			case 'v' : {
				if (strlen(argv[i])>2)
				{
					std::cerr<<"Unknown option: "<<argv[i]<<"\n";
					return false;
				}
				verbose = true;
				alignment::verbose = true;
				pairwise_cyclic_alignment::verbose = true;
				average_linkage_cluster::verbose = true;
				
				break;
			}
			case 'w' : {
				if (strlen(argv[i])>2)
				{
					std::cerr<<"Unknown option: "<<argv[i]<<"\n";
					return false;
				}
				tree_weights = true;
				break;
			}
			case 'p' : {
				if (strlen(argv[i])>2)
				{
					std::cerr<<"Unknown option: "<<argv[i]<<"\n";
					return false;
				}
				multiple = false;
				break;
			}
			case 'l' : {
				if (strlen(argv[i])>2)
				{
					std::cerr<<"Unknown option: "<<argv[i]<<"\n";
					return false;
				}
				linear_alignment = true;
				break;
			}
			case 'o' : {
				if (strlen(argv[i])>2)
				{
					std::cerr<<"Unknown option: "<<argv[i]<<"\n";
					return false;
				}
				local_alignment = true;
				break;
			}
			case 'H' : {
				if (strlen(argv[i])>2)
				{
					std::cerr<<"Unknown option: "<<argv[i]<<"\n";
					return false;
				}
				fast_heuristic = true;
				break;
			}
			case 'L' : {
				if (strlen(argv[i])>2)
				{
					std::cerr<<"Unknown option: "<<argv[i]<<"\n";
					return false;
				}
				if (argc<=i+1 || !is_int(argv[i+1]))
				{
					std::cerr<<"expected integers specifying width of header and body.\n";
					return false;
				}
				if (i+1<argc && argv[i+1][0]!='-')
				{
					header_width = atoi(argv[i+1]);
					i++;
					num_of_options++;
				} 
				if (argc<=i+1 || !is_int(argv[i+1]))
				{
					std::cerr<<"expected integers specifying width of header and body.\n";
					return false;
				}
				if (i+1<argc && argv[i+1][0]!='-')
				{
					body_width = atoi(argv[i+1]);
					i++;
					num_of_options++;
				}
				break;
			}
			case 'g' : {
				if (strlen(argv[i])>2)
				{
					std::cerr<<"Unknown option: "<<argv[i]<<"\n";
					return false;
				}
				gap_cost_changed = true;
				if (argc<=i+1 || !is_flt(argv[i+1]))
				{
					std::cerr<<"expected integers specifying width of header and body.\n";
					return false;
				}
				if (i+1<argc && argv[i+1][0]!='-')
				{
					gap_open_penalty = atof(argv[i+1]);
					i++;
					num_of_options++;
				} 
				if (argc<=i+1 || !is_flt(argv[i+1]))
				{
					std::cerr<<"expected integers specifying width of header and body.\n";
					return false;
				}
				if (i+1<argc && argv[i+1][0]!='-')
				{
					gap_ext_penalty = atof(argv[i+1]);
					i++;
					num_of_options++;
				}
				break;
			}
			case 'O' : {
				if (strlen(argv[i])>2)
				{
					std::cerr<<"Unknown option: "<<argv[i]<<"\n";
					return false;
				}
				if (argc<=i+1)
				{
					std::cerr<<"expected output filename.\n";
					return false;
				}
				if (i+1<argc && argv[i+1][0]!='-')
				{
					output_filename = std::string(argv[i+1]);
					i++;
					num_of_options++;
				}
				break;
			}
			case 's' : {
				if (strlen(argv[i])>2)
				{
					std::cerr<<"Unknown option: "<<argv[i]<<"\n";
					return false;
				}
				get_shifted_sequences = true;
				break;
			}
			case 'C' : {
				if (strlen(argv[i])>2)
				{
					std::cerr<<"Unknown option: "<<argv[i]<<"\n";
					return false;
				}
				clustalw_output = true;
				average_linkage_cluster::clustalw_output = true;
				break;
			}
			case 'j' : {
				if (strlen(argv[i])>2)
				{
					std::cerr<<"Unknown option: "<<argv[i]<<"\n";
					return false;
				}
				use_joint_match_score = true;
				pairwise_cyclic_alignment::use_joint_match_score = true;
				break;
			}
			default: {
				return false;
			}
		}
	}
	return true;
	
}



int main(int argc,char **argv)
{

	if (!read_parameters(argc,argv))
	{
		print_help_message();
		return 0;
	}

	std::ofstream output_stream;
	if (output_filename!="") {
		output_stream.open(output_filename.c_str(),std::ofstream::out);
		if (!output_stream.good()) {
			std::cout<<"could not open output stream.\n";
			return -1;
		}
	}
	if (verbose)
	{
		std::cout<<"input file: "<<argv[num_of_options]<<"\n";
		if (output_filename!="")
			std::cout<<"output file: "<<output_filename<<"\n";
	}
	fasta_reader fastafile (argv[num_of_options]);    // caveat arg list!
	std::string header_1,seq_1;
	std::string header_2,seq_2;

	try {
		if (!multiple) {
			if (fastafile.get_seq (header_1, seq_1)) {
				if (fastafile.get_seq (header_2, seq_2)) {
					sequence_type aln_seq_1(seq_1,header_1);
					sequence_type aln_seq_2(seq_2,header_2);
					pairwise_cyclic_alignment A_pairwise(aln_seq_1,aln_seq_2);
					A_pairwise.set_output_width(header_width, body_width);
					if (gap_cost_changed)
						A_pairwise.set_gap_costs(10.*gap_open_penalty,10.*gap_ext_penalty);
					if (linear_alignment) {
						std::cout<<"\ncomputing linear alignment\n\n";
						A_pairwise.run_linear();
					} else if (fast_heuristic) {
						std::cout<<"\ncomputing heuristic alignment\n\n";
						A_pairwise.run_heuristic();
						std::cout<<"trace score = "<<A_pairwise.trace_score<<"\n";
					} else if (local_alignment) {
						std::cout<<"\ncomputing local alignment\n\n";
						A_pairwise.run_local();
					} else {
						std::cout<<"\ncomputing cyclic alignment\n\n";
						A_pairwise.run();
					}
					if (get_shifted_sequences)
						std::cout<<"Option -s not supported in combination with -p; printing alignment.\n";
					if (output_filename=="")
						std::cout<<A_pairwise;
					else
						output_stream<<A_pairwise;
				}
			}
		} else {
			// read fasta sequences into vector
			sequence_vector all_sequences;
			std::string header, seq;
			while (fastafile.get_seq (header, seq)) {
				all_sequences.push_back(sequence_type(seq,header));
			}
			guide_tree_alignment alignment_tree(all_sequences);
			alignment_tree.set_weighted(tree_weights);
			if (gap_cost_changed)
				alignment_tree.set_gap_costs(gap_open_penalty,gap_ext_penalty);
			if (!tree_guided) {
				std::cout<<"\n computing agglomorative alignment\n\n";
				alignment_tree.run();
				alignment_tree.set_output_width(header_width, body_width);
			} else {
				std::cout<<"\n computing tree guided alignment\n\n";
				alignment_tree.run_tree_guided();
				alignment_tree.set_output_width(header_width, body_width);
			}
			if (output_filename=="")
				if (!get_shifted_sequences)
					std::cout<<alignment_tree;
				else
					std::cout<<alignment_tree.get_shifted_sequences();
			else
				if (!get_shifted_sequences)
					output_stream<<alignment_tree;
				else
					output_stream<<alignment_tree.get_shifted_sequences();
		}
	} catch(std::exception e) {
		std::cout<<"error:\n"<<e.what()<<"\n";
	}

}


