#include <algorithm>
#include "alignment.h"
#include <list>
#include <string>
#include <exception>


typedef std::list<std::string> string_list;
typedef alignment::float_list float_list;

bool alignment::verbose = false;

float distance(const alignment& A, const alignment& B)
{
	int L;
	if (A.L<B.L)
		L = A.L;
	else
		L = B.L;
	
	int i,j,k;
	int differences = 0;
	for (i=0; i<L; i++)
	{
		std::string A_column_i = A.columns[i];
		std::string B_column_i = B.columns[i];		
		for (j=0; j<A.K; j++)
			for (k=0; k<B.K; k++)
			{
				if (A_column_i[j]!=B_column_i[k])
					differences++;
			}
	}
	return ((float)(differences))/((float)L*A.K*B.K);
	
}

// =========================================================
bool is_gap(const std::string& A)
// =========================================================
{
	std::string::const_iterator A_it = A.begin();
	std::string::const_iterator A_end = A.end();
	for (; A_it!=A_end; ++A_it)
		if (*A_it != '-')
			return false;
	return true;
}

// =========================================================
alignment::alignment()
// =========================================================
{
	clear();
	header_width = 8;
	body_width = 60;
}

// =========================================================
alignment::~alignment()
// =========================================================
{
	clear();
}

// =========================================================
void alignment::clear()
// =========================================================
{
	columns.clear();
	line_names.clear();
	line_weights.clear();
}


// =========================================================
alignment::alignment(const std::string& seq)
// =========================================================
{
	K = 1;
	L = seq.size();
	header_width = 8;
	body_width = 60;
	
	columns.clear();
	columns.resize(L);

	line_names.clear();
	line_names.push_back("*");
	line_weights.clear();
	line_weights.push_back(1.);

	int j;
	for (j=0; j<L; j++)
	{
		columns[j].resize(1);
		columns[j][0] = seq[j];
	}
	
}

// =========================================================
alignment::alignment(const std::string& seq, const std::string& name_)
// =========================================================
{
	K = 1;
	L = seq.size();
	header_width = 8;
	body_width = 60;
	
	line_names.clear();
	line_names.push_back(name_);
	line_weights.clear();
	line_weights.push_back(1.);
	
	columns.clear();
	columns.resize(L);

	int j;
	for (j=0; j<L; j++)
	{
		columns[j].resize(1);
		columns[j][0] = seq[j];
	}
	
}

// initalize alignment to K sequences of length L
// =========================================================
alignment::alignment(int L_, int K_)
// =========================================================
{
	int i,j;
	L = L_;
	K = K_;
	header_width = 8;
	body_width = 60;
	
	columns.clear();
	string_vector sv(L);
	columns = sv;
	for (j=0; j<L; j++)
	{
		columns[j] = std::string(K,'*');
	}
	line_names.clear();
	line_weights.clear();
	for (i=0; i<K; i++)
	{
		line_names.push_back("*");
		line_weights.push_back(1.);
	}
	
}

// =========================================================
void alignment::merge(const alignment& aln)
// =========================================================
{
	int i;
	
	if (L!=aln.L)
		return;
	
	for (i=0; i<L; i++)
	{
		// concatenate column strings...
		columns[i].append(aln.columns[i]);
	}
	K += aln.K;
	string_list::const_iterator n_it = aln.line_names.begin();
	float_list::const_iterator w_it = aln.line_weights.begin();
	string_list::const_iterator n_end = aln.line_names.end();
	
	for (; n_it!=n_end; ++n_it, ++w_it)
	{
		line_names.push_back(*n_it);
		line_weights.push_back(*w_it);
	}
	
}

// =========================================================
int alignment::size() const
// =========================================================
{
	return columns.size();
}

// =========================================================
int alignment::height() const
// =========================================================
{
	return K;
}

// =========================================================
void alignment::set_weights(float w) 
// =========================================================
{
	float_list::iterator w_it = line_weights.begin();
	float_list::iterator w_end = line_weights.end();
	for (; w_it!=w_end; ++w_it)
		*w_it = w;
}


// =========================================================
void alignment::set_weights(const float_list& new_weights) 
// =========================================================
{
	line_weights = new_weights;
}

// =========================================================
void alignment::add_weight(float w) 
// =========================================================
{
	w /= (float)(height());
	float_list::iterator w_it = line_weights.begin();
	float_list::iterator w_end = line_weights.end();
	for (; w_it!=w_end; ++w_it)
	{
		*w_it += w;
	}		
}

// =========================================================
void alignment::add_weights(const float_list& added_weights) 
// =========================================================
{
	
	float_list::iterator w_it = line_weights.begin();
	float_list::iterator w_end = line_weights.end();
	float_list::const_iterator a_it = added_weights.begin();
	float_list::const_iterator a_end = added_weights.end();
	for (; w_it!=w_end, a_it!=a_end; ++w_it, ++a_it)
	{ 
		//std::cout<<*w_it<<" / 2 + "<<*a_it<<" = ";
		if (K==1)
			//*w_it /= (2.)*(float)K;
		//else {
			//std::cout<<"HUHU\n";
			*w_it = 0.;
		//}
		*w_it += *a_it;
		//std::cout<<*w_it<<"\n";
		//float db = *w_it;
		//float db2 = db;
	}		
}


// =========================================================
float_list alignment::get_weights()
// =========================================================
{
	return line_weights;
}

// =========================================================
float_list alignment::get_normalized_weights()
// =========================================================
{
	float max_w = 0.;
	float_list::iterator w_it = line_weights.begin();
	float_list::iterator w_end = line_weights.end();
	for (; w_it!=w_end; ++w_it)
	{
		if (*w_it>max_w)
			max_w = *w_it;
	}
	
	float_list result_list = line_weights;
	if (max_w==0.)
		return line_weights;

	w_it = result_list.begin();
	w_end = result_list.end();
	for (; w_it!=w_end; ++w_it)
		*w_it /= max_w;
	return result_list;
}


// =========================================================
void alignment::reverse()
// =========================================================
{
	std::reverse(columns.begin(),columns.end());
}

// =========================================================
void alignment::resize(int s)
// =========================================================
{
	return columns.resize(s);
}

//alignment alignment::subaln(int start_pos, int length)
//{
//	alignment result;
//	result.columns = string_vector(columns[start_pos], columns[start_pos]+length);
//	return result;
//}

// =========================================================
float alignment::get_distance()
// =========================================================
{
	int i,j,k;
	int differences = 0;
	for (i=0; i<L; i++)
	{
		std::string column_i = columns[i];		
		for (j=0; j<K; j++)
			for (k=j+1; k<K; k++)
			{
				if (column_i[j]!=column_i[k])
					differences++;
			}
	}
	return ((float)(2*differences))/((float)L*K*(K-1));
}

// =========================================================
float alignment::get_score()
// =========================================================
{
	return K*L;
}



// return the i-th sequence, i.e., the i-th row
// =========================================================
std::string alignment::operator()(int i)
// =========================================================
{
	std::string seq(L,'*');
	int j;
	for (j=0; j<L; j++)
	{
		seq[j] = columns[i][j];
	}
	return seq;
}

// return j-th entry in the i-th column
// =========================================================
char& alignment::operator()(int i, int j)
// =========================================================
{
	return columns[i][j];
}

// return the i-th column of the alignment
// =========================================================
std::string& alignment::operator[](int i)
// =========================================================
{
	return columns[i];
}


// =========================================================
void alignment::set_output_width(int header_w, int body_w)
// =========================================================
{
	header_width = header_w;
	body_width = body_w;
}

// =========================================================
std::string alignment::get_shifted_sequences()
// =========================================================
{
	std::string seq = "";
	int i,jj;
	
	string_list::const_iterator name_it = line_names.begin();
	
	for (i=0; i<K; ++i, ++name_it)
	{
		std::string name = *name_it;
		seq.push_back('>');
		seq.append(name);
		seq.push_back('\n');
		for (jj=0; jj<L; jj++)
			if (columns[jj][i] != '-')
				seq.push_back(columns[jj][i]);
		seq.push_back('\n');
	}
	std::string res = seq;
	return res;
}


// =========================================================
void alignment::print_weight_string()
// =========================================================
{
std::cout<<"\n\nsequence weights:\n";
	float_list::iterator w_it = line_weights.begin();
	float_list::iterator w_end = line_weights.end();
	string_list::iterator n_it = line_names.begin();
	string_list::iterator n_end = line_names.end();
	for (; w_it!=w_end && n_it!=n_end; ++w_it, ++n_it)
	{
		std::cout<<*n_it<<": "<<*w_it<<"\n";
	}
	std::cout<<"\n";
}

// =========================================================
std::ostream& operator<<(std::ostream& os, const alignment& B)
// =========================================================
{
	
	int i,j,jj;
	int b_width = B.body_width;

	if (B.body_width<=0)
		b_width = B.L+1;
		
	if (B.columns.size()<B.L)
	{
		throw std::exception();
		return os;
	}
	
	j=0;
	while (j<B.L)
	{
		string_list::const_iterator name_it = B.line_names.begin();
		if (B.columns[j].size()<B.K)
		{
			throw std::exception();
			return os;
		}

		for (i=0; i<B.K; i++, ++name_it)
		{
			std::string name = *name_it;
			bool unspaced = true;
			for (jj=0; jj<B.header_width; jj++)
			{
				if (jj<name.size() && unspaced) {
					os<<name[jj];
					if (name[jj]==' ')
						unspaced = false;
				} else {
					os<<" ";
				}
			}
			os<<" ";
			jj = j+b_width < B.L ? j+b_width : B.L;
			int end_j = j+b_width < B.L ? j+b_width : B.L;
			for (jj=j; jj<end_j; jj++)
				os<<B.columns[jj][i];
			os<<"\n";
		}
		j = jj;		
		os<<"\n";
	}
	return os;
}

