#ifndef ALIGNMENT
#define ALIGNMENT

#include <iostream>
#include <string>
#include <vector>
#include <list>

class alignment;

bool is_gap(const std::string& A);

float distance(const alignment& A, const alignment& B);

class alignment
{
	
public:

	typedef std::vector< std::string > string_vector;
	typedef std::list< std::string > string_list;
	typedef std::list< float > float_list;
	
	/// number of sequences (number of rows) in the alignment
	int K;

	/// length of the alignment (number of columns)
	int L;

	static bool verbose;
	
	/// the alignment is stored columnwise
	string_vector columns;

	/// lines correspond to sequences, each of which might want to carry
	/// its name along.
	string_list line_names;
		
	/// for taking into account weights for sequences such as in ClustalW
	float_list line_weights;

	/// number of header characters to be printed by operator<<
	int header_width;

	/// width of one single block as to be printed by operator<<
	int body_width;
	
	alignment();

	~alignment();

	void clear();

	/// number of columns in the alignment
	int size() const;

	/// number of rows in the alignment
	int height() const;

	void resize(int s);

	void reverse();

	void merge(const alignment&);

	void set_weights(float w);

	void set_weights(const float_list& new_weights);

	void add_weight(float w);

	void add_weights(const float_list& added_weights);

	float_list get_weights();	
	
	/// set the maximum weight to 1 and scale all other weights by a
	/// corresponding factor.
	float_list get_normalized_weights();
	
	/// compute the mean number of differences per residue, which can be
	/// used as a basis fdor a neighbour-joining guide tree for a multiple
	/// alignment in ClustalW manner.
	float get_distance();

	float get_score();

	void print_weight_string();
	
	void set_output_width(int header_w, int body_w);
	
	//alignment subaln(int start_pos, int length);

	/// construct a single-sequence alignment (without a name) from a string
	alignment(const std::string&);

	/// construct a single-sequence alignment (without a name) from a string
	alignment(const std::string&, const std::string& name);

	/// initalize alignment to K sequences of length L
	alignment(int L_, int K_);

	/// initalize alignment to an unspecified number of rows of length L
	alignment(int L_);

	/// return the j-th sequence, i.e., the j-th row
	std::string operator()(int j);

	/// return j-th entry in the i-th column
	char& operator()(int i, int j);

	/// return the i-th column of the alignment
	std::string& operator[](int i);

	std::string get_shifted_sequences();
	
	friend std::ostream& operator<<(std::ostream& os, const alignment& B);

};


#endif

