#ifndef COMPLETE_LINKAGE_CLUSTER
#define COMPLETE_LINKAGE_CLUSTER

#include <list>
#include <vector>
#include <limits.h>
#include <float.h>
#include <ostream>
#include <iostream>



// =========================================================	
template<class point>
class hierarchy_node
{
public:

	float diameter;

	int num_of_points;
	
	virtual bool is_leaf() {return true;};

	virtual void print(std::ostream& o) {o<<"VOID";}

	hierarchy_node() {diameter = 0.;}
	
	virtual ~hierarchy_node() {}

};
// =========================================================	


// =========================================================	
template<class point>
class internal_hierarchy_node : public hierarchy_node<point>
{
public:	

	hierarchy_node<point>* left;
	hierarchy_node<point>* right;
	
	internal_hierarchy_node(hierarchy_node<point>* node_1, hierarchy_node<point>* node_2, float diam)
	{
		left = node_1;
		right = node_2;
		diameter = diam;
		num_of_points = left->num_of_points + right->num_of_points;
	}

	bool is_leaf() {return false;}

	void print(std::ostream& o)
	{
		o<<"( ";
		left->print(o);
		o<<" , ";
		right->print(o);
		o<<" )";
	}

	~internal_hierarchy_node()
	{
		delete left;
		delete right;
	}
	
	void separate_largest_cluster(
		float diam,
		std::list< hierarchy_node<point>* >& trace,
		int& best_cardinality);
	
};
// =========================================================	


// =========================================================	
template<class point>
class hierarchy_leaf : public hierarchy_node<point>
{
	
public:
	
	point P;

	bool is_leaf() {return true;};
	
	hierarchy_leaf()
	{
		diameter = 0.;
	}
	
	hierarchy_leaf(const point& p)
	{
		num_of_points = 1;
		diameter = 0.;
		P = p;
	}

	void print(std::ostream& o)
	{
		o<<P;
	}

	~hierarchy_leaf() {}
	
};
// =========================================================	

// =========================================================
// =========================================================	
template<class point>
class complete_linkage_cluster
// =========================================================	
{
	
public:
	

	typedef typename std::list< hierarchy_node<point>* > hierarchy;
	typedef typename std::vector<point> point_vector;
	typedef typename std::list<point_vector> class_list;
	typedef typename hierarchy::iterator hierarchy_iterator;
	typedef typename class_list::iterator class_list_iterator;
	
	/// dimension of the vector space the points are drawn from
	int dim;

	/// number of points
	int N;
	
	/// points to be classified 
	point_vector points;
	
	complete_linkage_cluster();

	~complete_linkage_cluster();

	complete_linkage_cluster(const point_vector &new_points, int new_dim);

	hierarchy_node<point>* root;

	void init(const point_vector &new_points, int new_dim);

	void clear();
	
	void run();

	void print(std::ostream& o);
	
	static float joint_diameter(const point_vector& class_1, const point_vector& class_2);
	static void merge_hierarchies(hierarchy_iterator root_1, hierarchy_iterator root_2, hierarchy& roots, float diam);
	static void merge_classes(class_list_iterator class_1, class_list_iterator class_2, class_list& classes);

	/// cut the largest cluster whose diameter is at most \c diam out
	/// of the hierarchy and return it. Note that the responsibility
	/// for deleting the returned pointer is transferred to the calling
	/// function!
	hierarchy_node<point>* separate_largest_cluster(float diam);
	
};
// =========================================================
// =========================================================	


// =========================================================
template<class point>
void print_class(typename std::vector<point> v)
// =========================================================	  
{
	typename complete_linkage_cluster<point>::point_vector::iterator v_it = v.begin();
	typename complete_linkage_cluster<point>::point_vector::iterator v_end = v.end();
	for (; v_it!=v_end; ++v_it)
	{
		std::cout<<*v_it<<"; ";
	}
	
}


// =========================================================
template<class point>
void complete_linkage_cluster<point>::print(std::ostream& o)
// =========================================================
{
	root->print(o);
	o<<"\n";
}
	

// =========================================================
template<class point>
complete_linkage_cluster<point>::complete_linkage_cluster()
// =========================================================
{
	root = NULL;
	points.clear();
	dim = 0;
}

// =========================================================
template<class point>
complete_linkage_cluster<point>::~complete_linkage_cluster()
// =========================================================
{
	if (root!=NULL)
	{
		std::cerr<<"deleting root\n";
		delete root;
	}
}


// =========================================================
template<class point>
complete_linkage_cluster<point>::complete_linkage_cluster(const complete_linkage_cluster<point>::point_vector &new_points, int new_dim)
// =========================================================
{
	root = NULL;
	init(new_points, new_dim);
}

// =========================================================
template<class point>
void complete_linkage_cluster<point>::init(const complete_linkage_cluster<point>::point_vector &new_points, int new_dim)
// =========================================================
{
	clear();
	points = new_points;
	dim = new_dim;
	N = points.size();
}

// =========================================================
template<class point>
void complete_linkage_cluster<point>::clear()
// =========================================================
{
	
	if (root!=NULL)
	{
		root = NULL;
		delete root;
	}
	
}

// =========================================================
template<class point>
void complete_linkage_cluster<point>::run()
// =========================================================
{
	
	//hierarchy roots;
	hierarchy roots;
	class_list classes;
	
	int i;
	for (i=0; i<N; i++)
	{
		hierarchy_leaf<point>* leaf_i = new hierarchy_leaf<point>(points[i]);
		roots.push_back(leaf_i);
		point_vector class_i(1);
		class_i[0] = points[i];
		classes.push_back(class_i);
	}

	hierarchy_iterator root_1, root_2, best_root_1, best_root_2;
	class_list_iterator class_1, class_2, best_class_1, best_class_2;

	int size = N;
	
	for (; size>1; --size)
	{
		std::cerr<<"*";
		root_1 = roots.begin();
		class_1 = classes.begin();
		best_root_1 = roots.end();
		best_root_2 = roots.end();
		best_class_1 = classes.end();
		best_class_2 = classes.end();

		float best_diam = FLT_MAX;
		
		for (; root_1!=roots.end(); ++root_1, ++class_1)
		{
			root_2 = root_1;
			class_2 = class_1;
			root_2++;
			class_2++;
			for (; root_2!=roots.end(); ++root_2, ++class_2)
			{
				float diam = joint_diameter(*class_1,*class_2);
				if (diam<best_diam)
				{
					best_diam = diam;
					best_root_1 = root_1;
					best_root_2 = root_2;
					best_class_1 = class_1;
					best_class_2 = class_2;

					// ---
					/* 					std::cout<<"->"<<"diam "<<best_diam<<"\n->"; */
					/* 					(**best_root_1).print(std::cout); */
					/* 					std::cout<<"\n->"; */
					/* 					(**best_root_2).print(std::cout); */
					/* 					std::cout<<"\n->"; */
					/* 					print_class(*best_class_1); */
					/* 					std::cout<<"\n->"; */
					/* 					print_class(*best_class_2); */
					/* 					std::cout<<"\n"; */
					// ---
					
				}
			}
		}
		// ---
		/* 		std::cout<<"------\n"<<"diam "<<best_diam<<"\n"; */
		/* 		(**best_root_1).print(std::cout); */
		/* 		std::cout<<"\n"; */
		/* 		(**best_root_2).print(std::cout); */
		/* 		std::cout<<"\n--\n"; */
		// ---

		merge_hierarchies(best_root_1,best_root_2,roots,best_diam);
		merge_classes(best_class_1,best_class_2,classes);
		hierarchy_iterator h_it = roots.begin();
		hierarchy_iterator h_end = roots.end();

		// ---
		/* 		for (; h_it!=h_end; ++h_it) */
		/* 		{ */
		/* 			(**h_it).print(std::cout); */
		/* 			std::cout<<"\n"; */
		/* 		} */
		// ---
		
	}
	root = *roots.begin();
	std::cerr<<std::endl;
	separate_largest_cluster(1.);
	separate_largest_cluster(2.);	
	separate_largest_cluster(5.);
	separate_largest_cluster(10.);
	separate_largest_cluster(100.);	
}


// =========================================================
template <class point>
float complete_linkage_cluster<point>::joint_diameter(const point_vector& class_1, const point_vector& class_2)
// =========================================================
{
	int m = class_1.size();
	int n = class_2.size();
	int i,j;
	float diam = 0.0;
	
	for (i=0; i<m; i++)
			for (j=0; j<n; j++)
			{
				point class_1_i = class_1[i];
				point class_2_j = class_2[j];
				float dist = vector_distance(class_1_i,class_2_j);
				if (dist>diam)
					diam = dist;
			}
	return diam;
	
};


// =========================================================
template <class point>
void complete_linkage_cluster<point>::merge_hierarchies(complete_linkage_cluster<point>::hierarchy_iterator root_1, complete_linkage_cluster<point>::hierarchy_iterator root_2, complete_linkage_cluster<point>::hierarchy& roots, float diam)
// =========================================================	  
{
	
	internal_hierarchy_node<point>* node = new internal_hierarchy_node<point>(*root_1,*root_2,diam);
	roots.erase(root_1);
	roots.erase(root_2);
	roots.push_back(node);
}


// =========================================================
template <class point>
void complete_linkage_cluster<point>::merge_classes(complete_linkage_cluster<point>::class_list_iterator class_1, complete_linkage_cluster<point>::class_list_iterator class_2, complete_linkage_cluster<point>::class_list& classes)
// =========================================================	  
{
	// this implementation is still ugly and needs quite some improvement!
	point_vector new_class;
	
	typename point_vector::iterator cl_it = class_1->begin();
	typename point_vector::iterator cl_end = class_1->end();
	for (; cl_it != cl_end; ++cl_it)
		new_class.push_back(*cl_it);
	cl_it = class_2->begin();
	cl_end = class_2->end();
	for (; cl_it != cl_end; ++cl_it)
		new_class.push_back(*cl_it);
	classes.erase(class_1);
	classes.erase(class_2);
	classes.push_back(new_class);

}

/* // ========================================================= */
/* template <class point> */
/* void complete_linkage_cluster<point>::merge_classes(complete_linkage_cluster<point>::class_list_iterator class_1, complete_linkage_cluster<point>::class_list_iterator class_2, complete_linkage_cluster<point>::class_list& classes) */
/* // =========================================================	   */
/* { */
/* 	// this implementation is still ugly and needs quite some improvement! */
/* 	point_vector new_class; */
	
/* 	typename point_vector::iterator cl_it = class_1->begin(); */
/* 	typename point_vector::iterator cl_end = class_1->end(); */
/* 	for (; cl_it != cl_end; ++cl_it) */
/* 		new_class.push_back(*cl_it); */
/* 	cl_it = class_2->begin(); */
/* 	cl_end = class_2->end(); */
/* 	for (; cl_it != cl_end; ++cl_it) */
/* 		new_class.push_back(*cl_it); */
/* 	classes.erase(class_1); */
/* 	classes.erase(class_2); */
/* 	classes.push_back(new_class); */

/* } */

// =========================================================
template <class point>
hierarchy_node<point>* complete_linkage_cluster<point>::separate_largest_cluster(float diam)
// =========================================================
{
	if (root->diameter<=diam)
	{
		hierarchy_node<point>* res = root;
		root = NULL;
		return res;
	}
	hierarchy_node<point>* best_cluster = NULL;
	hierarchy_node<point>* parent_best_cluster = NULL;
	int cardinality;
	bool is_left_son;

	std::list< hierarchy_node<point>* > trace;
	trace.push_front(root);
	root->separate_largest_cluster(diam, trace, cardinality);
	std::cout<<"largest cluster with diameter <= "<<diam<<" contains "<<trace.begin()->cardinality<<" elements.\n";
	
	//update cardinalities along trace
	//typename std::list< hierarchy_node<point> >::iterator t_it = trace.begin();
	//typename std::list< hierarchy_node<point> >::iterator t_end = trace.end();
	//for (; t_it!=t_end; t_it++)
	//{
	//	t_it->cardinality -= cardinality;
	//}
}

// =========================================================
template <class point>
void internal_hierarchy_node<point>::separate_largest_cluster(
	float diam,
	std::list< hierarchy_node<point>* >& trace,
	int& best_cardinality)
// =========================================================
{
	// *this has already been rules out as being optimal, so that it is
	// left to the sons of this to yield optimal solutions.
	
	if (diam<0. || is_leaf())
	{
		// that dont make no sense. goodbye.
		return;
	}

	// examine left subtree
	int left_card = 0;
	std::list< hierarchy_node<point>* > left_trace;
	if (left!=NULL)
	{
		if (left->diameter<=diam)
		// since diam >= 0, and the diameter of a leaf is 0,
		// this condition will be satisfied whenver left is a leaf,
		// so that the recursion is guaranteed to terminate at leaf level.
		{
			left_trace.push_front(left);
			left_card = left->cardinality;
		} else {
			left->separate_largest_cluster(diam,left_trace,left_card);
		}
	}
	
	// examine right subtree
	int right_card = 0;
	std::list< hierarchy_node<point>* > right_trace;
	if (right!=NULL)
	{
		if (right->diameter<=diam)
		// since diam >= 0, and the diameter of a leaf is 0,
		// this condition will be satisfied whenver right is a leaf,
		// so that the recursion is guaranteed to terminate at leaf level.
		{
			right_trace.push_front(right);
			right_card = right->cardinality;
		} else {
			right->separate_largest_cluster(diam,right_trace,right_card);
		}
	}

	// who was better -- left or right?
	if (left_card>=right_card)
	{
		trace = left_trace;
		best_cardinality = left_card;
	} else {
		trace = right_trace;
		best_cardinality = right_card;
	}

	
}


	
#endif
