#ifndef PAIRWISE_CYCLIC_ALIGNEMNT
#define PAIRWISE_CYCLIC_ALIGNEMNT

#include <string>
#include <vector>
#include <exception>

#include "alignment.h"
#include "alignment_matrix.h"

class alignment_exception : public std::exception
{
	
public:
	
	int ID;

	alignment_exception(int id);

	virtual std::string what();
	
};


/// Implementation of the circular alignments with affine gap costs
/// based on Usama bin Gotoh's quadratic time algorithm for linear
/// alignment. Uses O(n^2m^2) time and O(nm) space in total.
class pairwise_cyclic_alignment
{

public:

	typedef int score_type;
	typedef alignment sequence_type;
	typedef std::list<float> float_list;

	static score_type SCORE_MIN;
	static bool	use_joint_match_score;
	static bool verbose;
	
	/// representative of cyclic string 1 for alignment
	sequence_type backup_X;
	/// representative of cyclic string 2 for alignment
	sequence_type backup_Y;

	/// representative of cyclic string 1 for alignment
	sequence_type X;
	/// representative of cyclic string 2 for alignment
	sequence_type Y;

	//std::string name_X;

	//std::string name_Y;
	
	/// length of X;
	int m;
	/// length of Y;
	int n;

	/// apply the (clustal-like) line_weights to the individual
	/// sequences?
	bool weighted;
	
	/// sequence w/ gaps X through the alignment path obtained by run() or run_linear()
	sequence_type trace_X;
	/// sequence w/ gaps Y through the alignment path obtained by run() or run_linear()	
	sequence_type trace_Y;

	/// result simply contains trace_X and trace_Y put together after DP and tracing.
	alignment result;

	/// output formatting parameter
	int header_width;

	/// output formatting parameter
	int body_width;

	/// Variable names follow the Gotoh paper in the following sense:
	/// variables named A_D, A_P, and A_Q denote the matrices D, P, and Q, respectively.
	
	/// alignment matrix containing the minimal cost between
	/// X_1...X_i and Y_1...Y_j at pos (i,j) in case X[i] matches w/ Y[j]
	alignment_matrix<score_type> A_M;
	/// A_H contains the minimum of A_M, A_P and A_P, at each pos (i,j)
	alignment_matrix<score_type> A_H;
	/// alignment matrix for matching sequence parts
	alignment_matrix<score_type> A_P;
	/// aligment matrix for gapped sequence parts
	alignment_matrix<score_type> A_Q;
	/// matrix for tracing back alignment paths through alignment matrix
	alignment_matrix<int> A_trace;

	/// gap open cost
	score_type gamma;
	/// gap extension cost
	score_type alpha;
	/// score for matching characters
	score_type single_match_score;
	/// score for mismatching characters	
	score_type single_mismatch_score;

	/// gap open cost for local alignments
	score_type local_gamma;
	/// gap extension cost for local alignments
	score_type local_alpha;
	/// score for matching characters for local alignments
	score_type single_local_match_score;
	/// score for mismatching characters for local alignments
	score_type single_local_mismatch_score;

	score_type trace_score;


	/// local alignment information
	int local_aln_start_X;
	int local_aln_start_Y;
	int local_aln_end_X;
	int local_aln_end_Y;

	/// information about the "core" of a local alignment
	int best_run_start_X;
	int best_run_start_Y;
	int best_run_end_X;
	int best_run_end_Y;
	score_type best_run_score;
	
	pairwise_cyclic_alignment();
	
	//~pairwise_cyclic_alignment();
	
	pairwise_cyclic_alignment(const sequence_type& X_, const sequence_type& Y_);

	score_type alignment_score(sequence_type aln1, sequence_type aln2);
	
	score_type alignment_score(sequence_type aln);
	
	inline score_type match_score(char x, char y);

   score_type match_score(const std::string& x, const std::string& y);
	
	score_type match_score(const std::string& x, const std::string& y, const float_list& w_x, const float_list& w_y);

	score_type joint_match_score(const std::string& x, const std::string& y);

	score_type joint_local_match_score(const std::string& x, const std::string& y);
	
	inline score_type local_match_score(char x, char y);

   inline score_type local_match_score(std::string& x, std::string& y);
	
	void set_gap_costs(int gamma_, int alpha_);

	void set_weighted(bool);

	void run();

	void run_linear();

	void run_local();

	void run_heuristic();

	void init_matrices();

	void init_match_start(int,int);

	void init_gap_start(int,int);

	void init_local_start(int,int);
	
	void compute_matrices(int k, int l);

	void compute_weighted_matrices(int k, int l);

	void compute_local_matrices(int k, int l);

	void trace_sequences(int k, int l, bool gap_start);

	void trace_weighted_sequences(int k, int l, bool gap_start);

	void trace_local(int I, int J);

	void get_best_pos(int& I, int& J);

	void set_output_width(int header_w, int body_w);

	friend std::ostream& operator<<(std::ostream&,pairwise_cyclic_alignment&);
};

/// compute a pairwise cyclic alignment of aln_1 and aln_2, put this
/// into result and return the score
float merge(alignment& aln_1, alignment& aln_2, alignment& result, bool weighted, pairwise_cyclic_alignment::score_type g_open, pairwise_cyclic_alignment::score_type g_ext, bool cost_change);

/// compute a pairwise cyclic alignment of aln_1 and aln_2, and return the score
float merge(const alignment& aln_1, const alignment& aln_2, bool weighted, pairwise_cyclic_alignment::score_type g_open, pairwise_cyclic_alignment::score_type g_ext, bool cost_change);

std::ostream& operator<<(std::ostream&,pairwise_cyclic_alignment&);

#endif



